import ast

from flake8_django.checkers import ModelDunderStrMissingChecker, ModelFieldChecker, ModelFormChecker, URLChecker, RenderChecker

__version__ = '0.0.3'


class DjangoStyleFinder(ast.NodeVisitor):
    """
    Visit the node, and return issues.
    """
    checkers = {
        'Call': [
            ModelFieldChecker(),
            URLChecker(),
            RenderChecker(),
        ],
        'ClassDef': [
            ModelFormChecker(),
            ModelDunderStrMissingChecker(),
        ]
    }

    def __init__(self, *args, **kwargs):
        super(DjangoStyleFinder, self).__init__(*args, **kwargs)
        self.issues = []

    def capture_issues_visitor(self, visitor, node):
        for checker in self.checkers[visitor]:
            issues = checker.run(node)
            if issues:
                self.issues.extend(issues)
        self.generic_visit(node)

    def visit_Call(self, node):
        self.capture_issues_visitor('Call', node)

    def visit_ClassDef(self, node):
        self.capture_issues_visitor('ClassDef', node)


class DjangoStyleChecker(object):
    """
    Check common Django Style errors
    """
    options = None
    name = 'flake8-django'
    version = __version__

    def __init__(self, tree, filename):
        self.tree = tree
        self.filename = filename

    def run(self):
        parser = DjangoStyleFinder()
        parser.visit(self.tree)

        for issue in parser.issues:
            yield (issue.lineno, issue.col, issue.message, DjangoStyleChecker)



from .model_dunder_str import ModelDunderStrMissingChecker
from .model_fields import ModelFieldChecker
from .model_form import ModelFormChecker
from .render import RenderChecker
from .urls import URLChecker


__all__ = ['ModelDunderStrMissingChecker', 'ModelFieldChecker', 'ModelFormChecker', 'RenderChecker', 'URLChecker']



import ast

from .checker import Checker
from .issue import Issue


class DJ03(Issue):
    code = 'DJ03'
    description = 'Use of locals() is not recommended in render function, use explicit arguments'


class RenderChecker(Checker):

    def run(self, node):
        """
        Captures the use of locals() in render function.
        """
        if self.get_call_name(node) != 'render':
            return
        issues = []
        for arg in node.args:
            if isinstance(arg, ast.Call) and arg.func.id == 'locals':
                issues.append(
                    DJ03(
                        lineno=node.lineno,
                        col=node.col_offset,
                    )
                )
        return issues



import ast

from .checker import Checker
from .issue import Issue


class DJ04(Issue):
    code = 'DJ04'
    description = 'not recommended to use dashes in url name, use underscore instead'


class DJ05(Issue):
    code = 'DJ05'
    description = 'Missing namespace in urls include()'


class URLChecker(Checker):
    """
    Checks for bad practices on url definition.
    """
    def __init__(self):
        super(URLChecker, self).__init__()
        self.checks = [self.capture_dash_in_url_name, self.capture_url_missing_namespace]

    def run(self, node):
        if self.get_call_name(node) != 'url':
            return
        issues = []
        for check in self.checks:
            issue = check(node)
            if issue:
                issues.append(issue)
        return issues

    def capture_dash_in_url_name(self, node):
        """
        Capture dash in URL name
        """
        for keyword in node.keywords:
            if keyword.arg == 'name' and '-' in keyword.value.s:
                return DJ04(
                    lineno=node.lineno,
                    col=node.col_offset,
                )

    def capture_url_missing_namespace(self, node):
        """
        Capture missing namespace in url include.
        """
        for arg in node.args:
            if not(isinstance(arg, ast.Call) and isinstance(arg.func, ast.Name)):
                continue
            if arg.func.id != 'include':
                continue
            for keyword in arg.keywords:
                if keyword.arg == 'namespace':
                    return

            return DJ05(
                lineno=node.lineno,
                col=node.col_offset,
            )



import ast

from .base_model_checker import BaseModelChecker
from .issue import Issue


class DJ06(Issue):
    code = 'DJ06'
    description = 'ModelForm.Meta should not set "exclude", set "fields" instead'


class DJ07(Issue):
    code = 'DJ07'
    description = "ModelForm.Meta should not set fields to '__all__'"


class ModelFormChecker(BaseModelChecker):
    model_name_lookup = 'ModelForm'

    def checker_applies(self, node):
        for base in node.bases:
            is_model_form = self.is_model_name_lookup(base) or self.is_models_name_lookup_attribute(base)
            if is_model_form:
                return True
        return False

    def is_string_dunder_all(self, element):
        """
        Return True if element is ast.Str or ast.Bytes and equals "__all__"
        """
        if not isinstance(element.value, (ast.Str, ast.Bytes)):
            return False

        node_value = element.value.s
        if isinstance(node_value, bytes):
            node_value = node_value.decode()
        return node_value == '__all__'

    def run(self, node):
        """
        Captures the use of exclude in ModelForm Meta
        """
        if not self.checker_applies(node):
            return

        issues = []
        for body in node.body:
            if not isinstance(body, ast.ClassDef):
                continue
            for element in body.body:
                if not isinstance(element, ast.Assign):
                    continue
                for target in element.targets:
                    if target.id == 'fields' and self.is_string_dunder_all(element):
                        issues.append(
                            DJ07(
                                lineno=node.lineno,
                                col=node.col_offset,
                            )
                        )
                    elif target.id == 'exclude':
                        issues.append(
                            DJ06(
                                lineno=node.lineno,
                                col=node.col_offset,
                            )
                        )
        return issues



from .checker import Checker
from .issue import Issue


NOT_NULL_TRUE_FIELDS = [
    'CharField', 'TextField', 'SlugField',
    'EmailField', 'UUIDField', 'ImageField',
    'FileField', 'FilePathField', 'URLField'
]
NOT_BLANK_TRUE_FIELDS = ['BooleanField']


class DJ01(Issue):
    code = 'DJ01'
    description = 'null=True not recommended to be used in {field}'


class DJ02(Issue):
    code = 'DJ02'
    description = 'blank=True not recommended to be used in {field}'


class ModelFieldChecker(Checker):

    def run(self, node):
        call_name = self.get_call_name(node)
        if not(call_name in NOT_NULL_TRUE_FIELDS or call_name in NOT_BLANK_TRUE_FIELDS):
            return

        issues = []
        for keyword in node.keywords:
            if call_name in NOT_NULL_TRUE_FIELDS and keyword.arg == 'null' and keyword.value.value is True:
                issues.append(
                    DJ01(
                        lineno=node.lineno,
                        col=node.col_offset,
                        parameters={'field': call_name}
                    )
                )
            if call_name in NOT_BLANK_TRUE_FIELDS and keyword.arg == 'blank' and keyword.value.value is True:
                issues.append(
                    DJ02(
                        lineno=node.lineno,
                        col=node.col_offset,
                        parameters={'field': call_name}
                    )
                )
        return issues



import ast

from .base_model_checker import BaseModelChecker
from .issue import Issue


class DJ08(Issue):
    code = 'DJ08'
    description = '__str__ method should be present in all db models'


class ModelDunderStrMissingChecker(BaseModelChecker):
    model_name_lookup = 'Model'

    def checker_applies(self, node):
        for base in node.bases:
            if self.is_model_name_lookup(base) or self.is_models_name_lookup_attribute(base):
                return True
        return False

    def is_dunder_str_method(self, element):
        return (
            isinstance(element, ast.FunctionDef) and
            element.name == '__str__'
        )

    def run(self, node):
        if not self.checker_applies(node):
            return

        if not any(self.is_dunder_str_method(elem) for elem in node.body):
            return [
                DJ08(
                    lineno=node.lineno,
                    col=node.col_offset
                )
            ]

        return []



import ast

from .checker import Checker


class BaseModelChecker(Checker):
    """
    Base class for checkers that must lookup for Model like nodes.
    """

    model_name_lookup = None

    def is_model_name_lookup(self, base):
        """
        Return True if class is defined as the respective model name lookup declaration
        """
        return (
            isinstance(base, ast.Name) and
            base.id == self.model_name_lookup
        )

    def is_models_name_lookup_attribute(self, base):
        """
        Return True if class is defined as the respective model name lookup declaration
        """
        return (
            isinstance(base, ast.Attribute) and
            isinstance(base.value, ast.Name) and
            base.value.id == 'models' and base.attr == self.model_name_lookup
        )



class Issue(object):
    """
    Abstract class for issues.
    """
    code = ''
    description = ''

    def __init__(self, lineno, col, parameters=None):
        self.parameters = {} if parameters is None else parameters
        self.col = col
        self.lineno = lineno

    @property
    def message(self):
        """
        Return issue message.
        """
        message = self.description.format(**self.parameters)
        return '{code} {message}'.format(code=self.code, message=message)



from os.path import dirname, join
from setuptools import find_packages, setup


with open(join(dirname(__file__), 'flake8_django/VERSION'), 'rb') as f:
    version = f.read().decode('ascii').strip()


with open('README.md', 'r') as f:
    long_description = f.read()


setup(
    name='flake8-django',
    license='GPL',
    version=version,
    description='Plugin to catch bad style specific to Django Projects',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Rocio Aramberri Schegel',
    author_email='rocioaramberri@schegel.net',
    url='http://github.com/rocioar/flake8-django',
    packages=find_packages(exclude=('tests', 'tests.*')),
    include_package_data=True,
    entry_points={
        'flake8.extension': [
            'DJ0 = flake8_django:DjangoStyleChecker',
        ],
    },
    install_requires=['flake8'],
    tests_require=['pytest'],
    classifiers=[
        'Framework :: Flake8',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Software Development :: Quality Assurance',
    ],
)



import ast


class Checker(object):
    """
    Abstract class for Checkers.
    """

    def get_call_name(self, node):
        """
        Return call name for the given node.
        """
        if isinstance(node.func, ast.Attribute):
            return node.func.attr
        elif isinstance(node.func, ast.Name):
            return node.func.id

    def run(self, node):
        """
        Method that runs the checks and returns the issues.
        """
        return NotImplementedError  # pragma: no cover
